"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.GrandAgent = void 0;
const bayLog_1 = require("../bayLog");
const acceptHandler_1 = require("./acceptHandler");
const nonBlockingHandler_1 = require("./nonBlockingHandler");
const memUsage_1 = require("../memUsage");
const commandReceiver_1 = require("./commandReceiver");
const bayserver_1 = require("../bayserver");
const process_1 = require("process");
const child_process_1 = require("child_process");
const arrayUtil_1 = require("../util/arrayUtil");
class GrandAgent {
    constructor(agentId, maxShips, anchorable) {
        this.aborted = false;
        this.timerId = null;
        this.interval_sec = 10;
        this.agentId = agentId;
        this.timerHandlers = [];
        if (anchorable) {
            this.acceptHandler = new acceptHandler_1.AcceptHandler(this, GrandAgent.anchorablePortMap);
        }
        this.anchorable = anchorable;
        this.nonBlockingHandler = new nonBlockingHandler_1.NonBlockingHandler(this);
        this.maxInboundShips = maxShips;
        this.startTimer();
    }
    toString() {
        return "agt#" + this.agentId;
    }
    startTimer() {
        this.timerId = setInterval(() => {
            try {
                for (const h of this.timerHandlers) {
                    h.onTimer();
                }
            }
            catch (e) {
                this.abort(e);
            }
        }, this.interval_sec * 1000);
    }
    stopTimer() {
        if (this.timerId !== null) {
            clearInterval(this.timerId);
            this.timerId = null;
        }
    }
    reloadCert() {
    }
    printUsage() {
        // print memory usage
        bayLog_1.BayLog.info("%s MemUsage", this);
        bayLog_1.BayLog.info("  Node.js information: %s", JSON.stringify(process.versions, null, 2));
        const memoryUsage = process.memoryUsage();
        bayLog_1.BayLog.info("  HeapTotal: %s Bytes", memoryUsage.heapTotal.toLocaleString('en-US'));
        bayLog_1.BayLog.info("  HeapUsed: %s Bytes", memoryUsage.heapUsed.toLocaleString('en-US'));
        bayLog_1.BayLog.info("  External: %s Bytes", memoryUsage.external.toLocaleString('en-US'));
        memUsage_1.MemUsage.get(this.agentId).printUsage(1);
    }
    shutdown() {
        bayLog_1.BayLog.debug("%s shutdown", this);
        if (this.acceptHandler != null)
            this.acceptHandler.shutdown();
        this.abort(null, 0);
    }
    abort(err = null, status = 1) {
        if (err) {
            bayLog_1.BayLog.fatal("%s abort", this);
            bayLog_1.BayLog.fatal_e(err);
        }
        if (this.aborted)
            return;
        this.stopTimer();
        this.commandReceiver.end();
        for (const lis of GrandAgent.listeners)
            lis.remove(this.agentId);
        GrandAgent.agents.delete(this.agentId);
        this.aborted = true;
        if (bayserver_1.BayServer.harbor.isMultiCore())
            setTimeout(() => {
                (0, process_1.exit)(status);
            }, 5000);
        else
            this.clean();
    }
    fork(agentId, monitorPort) {
        let newArgv = [];
        for (const arg of bayserver_1.BayServer.scriptArgs) {
            if (!arg.startsWith("-agentid=") &&
                !arg.startsWith("-monitorPort=") &&
                !arg.startsWith("-openPort"))
                newArgv.push(arg);
        }
        newArgv.push("-agentid=" + agentId);
        newArgv.push("-monitorPort=" + monitorPort);
        let child = (0, child_process_1.fork)(bayserver_1.BayServer.scriptName, newArgv);
        bayLog_1.BayLog.debug("agt#%d Child process spawned: child agt#%d pid=%d", this.agentId, agentId, child.pid);
        child.on("exit", (code, signal) => {
            bayLog_1.BayLog.debug("agt#%d child agt#%d is exited: code=%d sig=%d", this.agentId, agentId, code, signal);
        });
        for (const portDkr of GrandAgent.anchorablePortMap) {
            //BayLog.debug("agt#%d send server socket: child agt#%d port=%d", this.agentId, agentId, portDkr.docker.getPort())
            child.send("serverSocket", portDkr.ch);
        }
    }
    addTimerHandler(handler) {
        this.timerHandlers.push(handler);
    }
    removeTimerHandler(handler) {
        arrayUtil_1.ArrayUtil.remove(handler, this.timerHandlers);
    }
    clean() {
        this.nonBlockingHandler.closeAll();
        this.agentId = -1;
    }
    //////////////////////////////////////////////////////
    // Static methods
    //////////////////////////////////////////////////////
    static init(agtIds, anchorablePortMap, unanchorablePortMap, maxShips, multiCore) {
        this.agentCount = agtIds.length;
        this.maxShips = maxShips;
        this.multiCore = multiCore;
        this.anchorablePortMap = anchorablePortMap;
        this.agents = new Map();
        for (const agtId of agtIds) {
            this.add(agtId, true);
        }
    }
    static get(agentId) {
        return this.agents.get(agentId);
    }
    static getByIndex(idx) {
        let agents = Array.from(this.agents.values());
        return agents[idx];
    }
    static add(agentId, anchorable) {
        if (agentId == -1)
            agentId = this.maxAgentId + 1;
        //BayLog.debug("Add agent: id=%d", agentId);
        if (agentId > this.maxAgentId)
            this.maxAgentId = agentId;
        // Agents run on multi core (process mode)
        let agt = new GrandAgent(agentId, this.maxShips, anchorable);
        this.agents.set(agentId, agt);
        for (const lis of this.listeners) {
            lis.add(agentId);
        }
    }
    static addLifecycleListener(listener) {
        this.listeners.push(listener);
    }
    //////////////////////////////////////////////////////
    // Other methods
    //////////////////////////////////////////////////////
    runCommandReceiver(commCh) {
        this.commandReceiver = new commandReceiver_1.CommandReceiver(this, commCh);
    }
}
exports.GrandAgent = GrandAgent;
GrandAgent.SELECT_TIMEOUT_SEC = 10;
GrandAgent.CMD_OK = 0;
GrandAgent.CMD_CLOSE = 1;
GrandAgent.CMD_RELOAD_CERT = 2;
GrandAgent.CMD_MEM_USAGE = 3;
GrandAgent.CMD_SHUTDOWN = 4;
GrandAgent.CMD_ABORT = 5;
GrandAgent.CMD_FORK = 6;
GrandAgent.maxAgentId = 0;
GrandAgent.agents = null;
GrandAgent.listeners = [];
GrandAgent.INTERVAL_SEC = 10;
//# sourceMappingURL=grandAgent.js.map