"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.GrandAgentMonitor = void 0;
const child_process_1 = require("child_process");
const bayLog_1 = require("../bayLog");
const bayserver_1 = require("../bayserver");
const ioUtil_1 = require("../util/ioUtil");
const grandAgent_1 = require("./grandAgent");
const bayMessage_1 = require("../bayMessage");
const symbol_1 = require("../symbol");
const process_1 = require("process");
const net_1 = require("net");
class GrandAgentMonitor {
    constructor(agentId, anchorable, comChannel) {
        this.agentId = agentId;
        this.anchorable = anchorable;
        this.communicationChannel = comChannel;
        this.communicationChannel.on("data", (buf) => {
            this.onRead(buf);
        });
        this.communicationChannel.on("end", () => {
            bayLog_1.BayLog.info("%s Communication lost: agt#%d", this, agentId);
            GrandAgentMonitor.agentAborted(agentId, anchorable);
        });
        this.communicationChannel.on("error", (e) => {
            bayLog_1.BayLog.error_e(e, "%s Communication error", this);
            GrandAgentMonitor.agentAborted(this.agentId, this.anchorable);
        });
    }
    toString() {
        return "Monitor#" + this.agentId;
    }
    onRead(buf) {
        let pos = 0;
        while (pos < buf.length) {
            let close = false;
            if (buf.length == 0)
                close = true;
            else {
                let res = ioUtil_1.IOUtil.readInt64(buf, pos);
                let cmd = Number(res >> BigInt(48));
                if (cmd == grandAgent_1.GrandAgent.CMD_CLOSE)
                    close = true;
            }
            if (close) {
                bayLog_1.BayLog.debug("%s read Close", this);
                GrandAgentMonitor.agentAborted(this.agentId, this.anchorable);
            }
            else {
                bayLog_1.BayLog.debug("%s read OK", this);
            }
            pos += 8;
        }
    }
    reloadCert() {
        bayLog_1.BayLog.debug("%s send reload command", this);
        this.send(grandAgent_1.GrandAgent.CMD_RELOAD_CERT);
    }
    printUsage() {
        bayLog_1.BayLog.debug("%s send mem_usage command", this);
        this.send(grandAgent_1.GrandAgent.CMD_MEM_USAGE);
    }
    shutdown() {
        bayLog_1.BayLog.debug("%s send shutdown command", this);
        this.send(grandAgent_1.GrandAgent.CMD_SHUTDOWN);
    }
    abort() {
        bayLog_1.BayLog.debug("%s send abort command", this);
        try {
            this.send(grandAgent_1.GrandAgent.CMD_ABORT);
        }
        catch (e) {
            bayLog_1.BayLog.error_e(e);
        }
    }
    send(cmd, arg1 = 0, arg2 = 0, arg3 = 0) {
        bayLog_1.BayLog.debug("%s send command %d: arg=(%d, %d, %d)", this, cmd, arg1, arg2, arg3);
        let data64 = (BigInt(cmd) << BigInt(48) & BigInt(0xffff000000000000)) |
            (BigInt(arg1) << BigInt(32) & BigInt(0xffff00000000)) |
            (BigInt(arg2) << BigInt(16) & BigInt(0xffff0000)) |
            (BigInt(arg3) & BigInt(0xffff));
        ioUtil_1.IOUtil.writeInt64(this.communicationChannel, data64);
    }
    close() {
        this.communicationChannel.end();
    }
    //////////////////////////////////////////////////////
    // Class methods
    //////////////////////////////////////////////////////
    static init(numAgents, anchoredPortMap) {
        bayLog_1.BayLog.debug("Init GrandAgentMonitor numAgents=%d", numAgents);
        this.numAgents = numAgents;
        this.anchoredPortMap = anchoredPortMap;
        this.monitors = new Map();
        this.add(true, true, () => {
            for (let i = 1; i < this.numAgents; i++) {
                this.add(true, false, null);
            }
        });
    }
    static add(anchorable, openSocket, callback) {
        let agentId = this.curId++;
        bayLog_1.BayLog.debug("GrandAgentMonitor::Add agent id=%d", agentId);
        if (bayserver_1.BayServer.harbor.isMultiCore()) {
            let serverSkt = (0, net_1.createServer)((skt) => {
                this.monitors.set(agentId, new GrandAgentMonitor(agentId, anchorable, skt));
                //BayLog.debug("Connected from agt#%d", agentId)
                bayLog_1.BayLog.trace(bayMessage_1.BayMessage.get(symbol_1.Symbol.MSG_CLOSING_LOCAL_PORT, serverSkt.address().port));
                serverSkt.close();
                if (callback != null) {
                    callback();
                }
            });
            serverSkt.listen(0, () => {
                const port = serverSkt.address().port;
                bayLog_1.BayLog.trace("open port for child agt#%d: %d", agentId, port);
                if (openSocket) {
                    let newArgv = [...bayserver_1.BayServer.scriptArgs];
                    newArgv.push("-agentid=" + agentId);
                    newArgv.push("-monitorPort=" + port);
                    newArgv.push("-openPort");
                    let child = (0, child_process_1.fork)(bayserver_1.BayServer.scriptName, newArgv);
                    bayLog_1.BayLog.debug("Child process spawned: child agt#%d pid=%d", agentId, child.pid);
                    child.on("exit", (code) => {
                        bayLog_1.BayLog.debug("agt#%d is exited: %d", agentId, code);
                        this.agentAborted(agentId, anchorable);
                    });
                }
                else {
                    for (const [_, mon] of this.monitors) {
                        bayLog_1.BayLog.debug("Send fork command: to agt#%d, child agt#%d", mon.agentId, agentId);
                        mon.send(grandAgent_1.GrandAgent.CMD_FORK, agentId, port);
                        break;
                    }
                }
            });
        }
        else {
            for (const portDkr of this.anchoredPortMap) {
                ioUtil_1.IOUtil.openLocalPipe((serverSideSocket) => {
                    let agt = grandAgent_1.GrandAgent.get(agentId);
                    agt.runCommandReceiver(serverSideSocket);
                }, (clientSideSocket) => {
                    this.monitors.set(agentId, new GrandAgentMonitor(agentId, anchorable, clientSideSocket));
                });
            }
        }
    }
    static printUsageAll() {
        for (const mon of this.monitors.values()) {
            try {
                mon.printUsage();
            }
            catch (e) {
                bayLog_1.BayLog.error_e(e);
            }
        }
    }
    static reloadCertAll() {
        bayLog_1.BayLog.debug("Reload all");
        for (const mon of this.monitors.values()) {
            mon.reloadCert();
        }
    }
    static restartAll() {
        bayLog_1.BayLog.debug("Restart all");
        for (const mon of this.monitors.values())
            mon.shutdown();
    }
    static shutdownAll() {
        bayLog_1.BayLog.debug("Shutdown all");
        this.finale = true;
        for (const mon of this.monitors.values())
            mon.shutdown();
        setTimeout(() => {
            (0, process_1.exit)(0);
        }, this.AGENT_SHUTDOWN_WAIT);
    }
    static abortAll() {
        bayLog_1.BayLog.debug("Shutdown all");
        this.finale = true;
        for (const mon of this.monitors.values()) {
            mon.abort();
        }
        setTimeout(() => {
            (0, process_1.exit)(1);
        }, this.AGENT_SHUTDOWN_WAIT);
    }
    static agentAborted(agtId, anchorable) {
        let mon = this.monitors.get(agtId);
        if (mon == null) {
            // already handled
            return;
        }
        bayLog_1.BayLog.info(bayMessage_1.BayMessage.get(symbol_1.Symbol.MSG_GRAND_AGENT_SHUTDOWN, agtId));
        this.monitors.delete(agtId);
        if (!this.finale) {
            if (this.monitors.size < this.numAgents) {
                try {
                    if (!bayserver_1.BayServer.harbor.isMultiCore()) {
                        grandAgent_1.GrandAgent.add(-1, anchorable);
                    }
                    this.add(anchorable, false, null);
                }
                catch (e) {
                    bayLog_1.BayLog.error_e(e);
                }
            }
        }
    }
}
exports.GrandAgentMonitor = GrandAgentMonitor;
GrandAgentMonitor.AGENT_SHUTDOWN_WAIT = 3000;
GrandAgentMonitor.curId = 1;
GrandAgentMonitor.monitors = null;
//# sourceMappingURL=grandAgentMonitor.js.map