"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.NonBlockingHandler = void 0;
const bayLog_1 = require("../bayLog");
const nextSocketAction_1 = require("./nextSocketAction");
const buffer_1 = require("buffer");
const ioException_1 = require("../util/ioException");
const sink_1 = require("../sink");
const arrayUtil_1 = require("../util/arrayUtil");
const channelWrapper_1 = require("./channelWrapper");
const tls = require("tls");
const fs = require("fs");
class ChannelState {
    constructor(ch, listener) {
        this.ch = ch;
        this.listener = listener;
        this.accepted = false;
        this.closing = false;
    }
    access() {
        this.lastAccessTime = Date.now();
    }
    toString() {
        let s = "";
        if (this.listener)
            s = this.listener.toString();
        if (this.closing)
            s += " closing=true";
        return s;
    }
}
class NonBlockingHandler {
    constructor(agent) {
        this.chStates = [];
        this.chCount = 0;
        this.agent = agent;
        this.agent.addTimerHandler(this);
    }
    //////////////////////////////////////////////////////
    // Implements TimerHandler
    //////////////////////////////////////////////////////
    onTimer() {
        this.closeTimeoutSockets();
    }
    //////////////////////////////////////////////////////
    // Custom methods
    //////////////////////////////////////////////////////
    addChannelListener(ch, chLis) {
        let chState = new ChannelState(ch, chLis);
        this.addChannelState(ch, chState);
        if (ch.type == channelWrapper_1.ChannelWrapper.TYPE_SOCKET || ch.type == channelWrapper_1.ChannelWrapper.TYPE_READABLE) {
            this.setupSocketEvents(ch, chState);
        }
        chState.access();
        return chState;
    }
    askToStart(ch) {
        let chState = this.findChannelState(ch);
        bayLog_1.BayLog.debug("%s askToStart ch=%s", this.agent, chState);
        chState.accepted = true;
    }
    askToConnect(ch, addr, port) {
        if (ch == null)
            throw new Error("nullPo");
        bayLog_1.BayLog.debug("%s askToConnect addr=%s port=%d", this.agent, addr, port);
        let chState = this.findChannelState(ch);
        let callback = () => {
            try {
                this.onConnect(chState);
            }
            catch (e) {
                this.agent.abort(e);
            }
        };
        if (port == 0) {
            // UNIX domain socket
            ch.socket.connect(addr, callback);
        }
        else {
            ch.socket.connect(port, addr, callback);
        }
        //this.setupSocketEvents(ch, chState)
    }
    askToRead(ch) {
        bayLog_1.BayLog.debug("%s askToRead", this.agent);
        let chState = this.findChannelState(ch);
        if (chState == null) {
            bayLog_1.BayLog.warn("%s channel state not found: %s", this.agent, ch);
            return;
        }
        if (ch.type == channelWrapper_1.ChannelWrapper.TYPE_SOCKET) {
            if (ch.socket.isPaused())
                ch.socket.resume(); // read on
        }
        else if (ch.type == channelWrapper_1.ChannelWrapper.TYPE_READABLE) {
            if (ch.readable.isPaused())
                ch.readable.resume(); // read on
        }
        else {
            this.readFs(chState);
        }
        chState.access();
    }
    askToWrite(buf, ch, listener) {
        bayLog_1.BayLog.debug("%s askToWrite: len=%d", this.agent, buf.length);
        let chState = this.findChannelState(ch);
        if (chState == null) {
            bayLog_1.BayLog.warn("%s channel state not found: %s", this.agent, ch);
            return;
        }
        if (ch.type == channelWrapper_1.ChannelWrapper.TYPE_SOCKET) {
            ch.socket.write(buf, (err) => {
                if (err) {
                    bayLog_1.BayLog.debug("%s Write error on writing %d bytes", this.agent, buf.length);
                    // This error will be handled by error event such as ch.on("error", ...)
                    //    this.onError(chState, err)
                }
                else {
                    bayLog_1.BayLog.debug("%s Wrote: len=%d", this.agent, buf.length);
                }
                if (this.agent.aborted)
                    return;
                let chState = this.findChannelState(ch);
                if (chState == null || chState.closing) {
                    // channel is already closed
                    bayLog_1.BayLog.debug("Channel is closed: buflen=%d", buf.length);
                    return;
                }
                try {
                    this.onWrite(chState, listener);
                }
                catch (e) {
                    this.agent.abort(e);
                }
            });
        }
        else if (ch.type == channelWrapper_1.ChannelWrapper.TYPE_READABLE) {
            throw new sink_1.Sink();
        }
        else {
            fs.write(ch.fd, buf, 0, buf.length, (err, written) => {
                let chState = this.findChannelState(ch);
                try {
                    if (err) {
                        this.onError(chState, err);
                    }
                    else {
                        this.onWrite(chState, listener);
                    }
                }
                catch (e) {
                    this.agent.abort(e);
                }
            });
        }
        chState.access();
    }
    askToClose(ch) {
        let chState = this.findChannelState(ch);
        bayLog_1.BayLog.debug("%s askToClose chState=%s", this.agent, chState);
        if (chState == null) {
            bayLog_1.BayLog.warn("%s channel state not found: %s", this.agent, ch);
            return;
        }
        this.closeChannel(chState);
        chState.access();
    }
    closeTimeoutSockets() {
        if (this.chStates.length == 0)
            return;
        let closeList = [];
        let now = Date.now();
        for (const chState of this.chStates) {
            if (chState.listener != null) {
                try {
                    let durationSec = Math.floor((now - chState.lastAccessTime) / 1000);
                    if (chState.listener.checkTimeout(chState.ch, durationSec)) {
                        bayLog_1.BayLog.debug("%s timed out channel found: ch=%s", this.agent, chState.ch);
                        closeList.push(chState);
                    }
                }
                catch (e) {
                    if (!(e instanceof ioException_1.IOException))
                        throw e;
                    else {
                        bayLog_1.BayLog.error_e(e);
                        closeList.push(chState);
                    }
                }
            }
        }
        for (const chState of closeList) {
            this.closeChannel(chState);
        }
    }
    closeAll() {
        for (const chState of this.chStates) {
            this.closeChannel(chState);
        }
    }
    //////////////////////////////////////////////////////
    // Private methods
    //////////////////////////////////////////////////////
    addChannelState(ch, chState) {
        //BayLog.debug("%s add ch %s chState=%s", this.agent, ch, chState);
        for (let st of this.chStates) {
            if (st.ch === ch) {
                bayLog_1.BayLog.debug("%s already added", this.agent);
                return;
            }
        }
        this.chStates.push(chState);
        this.chCount++;
    }
    removeChannelState(ch) {
        //BayLog.debug("%s remove ch %s", this.agent, ch);
        for (let i = 0; i < this.chStates.length; i++) {
            if (this.chStates[i].ch === ch) {
                arrayUtil_1.ArrayUtil.removeAt(i, this.chStates);
                break;
            }
        }
        this.chCount--;
    }
    findChannelState(ch) {
        //BayLog.debug("%s find chState %s", this.agent, ch);
        for (const chState of this.chStates) {
            //BayLog.debug("%s EQ Check ch %s", this.agent, chState.ch);
            if (chState.ch === ch) {
                //BayLog.debug("%s find ch %s chState=%s", this.agent, ch, chState);
                return chState;
            }
        }
        return null;
    }
    setupSocketEvents(ch, chState) {
        ch.socket.on("error", (err) => {
            if (this.agent.aborted)
                return;
            let chState = this.findChannelState(ch);
            if (chState == null || chState.closing) {
                // channel is already closed
                bayLog_1.BayLog.debug_e(err, "%s Error occured after channel is closed", this.agent);
                return;
            }
            try {
                this.onError(chState, err);
            }
            catch (e) {
                this.agent.abort(e);
            }
        });
        ch.socket.on('data', (buf) => {
            if (this.agent.aborted)
                return;
            let chState = this.findChannelState(ch);
            if (chState == null || chState.closing) {
                // channel is already closed
                bayLog_1.BayLog.debug("%s Channel is already closed: buflen=%d", this.agent, buf.length);
                return;
            }
            try {
                this.onRead(buf, chState);
            }
            catch (e) {
                this.agent.abort(e);
            }
        });
        ch.socket.pause(); // read off
        if (ch.socket instanceof tls.TLSSocket)
            ch.socket.on("secureConnect", () => {
                bayLog_1.BayLog.debug("%s Secure connect", this.agent);
            });
        ch.socket.on('end', () => {
            if (this.agent.aborted)
                return;
            let chState = this.findChannelState(ch);
            if (chState == null || chState.closing) {
                // channel is already closed
                return;
            }
            bayLog_1.BayLog.debug("%s channel end event: %s", this.agent, chState);
            try {
                this.onEof(chState);
            }
            catch (e) {
                this.agent.abort(e);
            }
            chState.closing = true;
        });
    }
    onConnect(chState) {
        var nextAct;
        try {
            nextAct = chState.listener.onConnect(chState.ch);
        }
        catch (e) {
            if (e instanceof ioException_1.IOException) {
                bayLog_1.BayLog.error_e(e);
                nextAct = nextSocketAction_1.NextSocketAction.CLOSE;
            }
            else
                throw e;
        }
        if (nextAct == nextSocketAction_1.NextSocketAction.CONTINUE)
            this.askToRead(chState.ch);
        this.doNextAction(nextAct, chState, false);
    }
    onRead(buf, chState) {
        var nextAct;
        try {
            //BayLog.info("Read buf: len=%d blen=%d cont=%s", buf.length, buf.byteLength, buf)
            nextAct = chState.listener.onRead(chState.ch, buf);
        }
        catch (e) {
            if (e instanceof ioException_1.IOException) {
                bayLog_1.BayLog.error_e(e, "Error in listener.onRead()");
                nextAct = nextSocketAction_1.NextSocketAction.CLOSE;
            }
            else
                throw e;
        }
        this.doNextAction(nextAct, chState, true);
    }
    onWrite(chState, dataLis) {
        let nextAct;
        try {
            nextAct = chState.listener.onWrite(chState.ch, dataLis);
        }
        catch (e) {
            if (e instanceof ioException_1.IOException) {
                bayLog_1.BayLog.error_e(e);
                nextAct = nextSocketAction_1.NextSocketAction.CLOSE;
            }
            else
                throw e;
        }
        this.doNextAction(nextAct, chState, false);
    }
    onEof(chState) {
        let nextAct;
        try {
            nextAct = chState.listener.onRead(chState.ch, buffer_1.Buffer.alloc(0));
        }
        catch (e) {
            if (e instanceof ioException_1.IOException) {
                bayLog_1.BayLog.error_e(e);
            }
            else
                throw e;
        }
        this.doNextAction(nextAct, chState, false);
    }
    onError(chState, err) {
        chState.listener.onError(chState.ch, err);
        this.doNextAction(nextSocketAction_1.NextSocketAction.CLOSE, chState, false);
    }
    onClosed(chState) {
        chState.listener.onClosed(chState.ch);
    }
    doNextAction(act, chState, isReading) {
        switch (act) {
            case nextSocketAction_1.NextSocketAction.CONTINUE:
                if (!isReading)
                    break;
            case nextSocketAction_1.NextSocketAction.READ:
                if (!(chState.ch.type == channelWrapper_1.ChannelWrapper.TYPE_SOCKET || chState.ch.type == channelWrapper_1.ChannelWrapper.TYPE_READABLE)) {
                    this.readFs(chState);
                }
                break;
            case nextSocketAction_1.NextSocketAction.WRITE:
                break;
            case nextSocketAction_1.NextSocketAction.CLOSE:
                this.closeChannel(chState);
                break;
            case nextSocketAction_1.NextSocketAction.SUSPEND:
                if (chState.ch.type == channelWrapper_1.ChannelWrapper.TYPE_SOCKET) {
                    bayLog_1.BayLog.debug("%s Suspend socket reading", this.agent);
                    chState.ch.socket.pause();
                }
                else if (chState.ch.type == channelWrapper_1.ChannelWrapper.TYPE_READABLE) {
                    bayLog_1.BayLog.debug("%s Suspend readable reading", this.agent);
                    chState.ch.readable.pause();
                }
                break;
        }
        chState.access();
    }
    readFs(chState) {
        let buf = chState.listener.getBuf();
        fs.read(chState.ch.fd, buf, 0, buf.length, -1, (err, bytesRead, buffer) => {
            if (err) {
                this.onError(chState, err);
            }
            else {
                try {
                    //if(bytesRead == 0)
                    //    return // no data read
                    if (bytesRead < buf.length) {
                        let newBuf = buffer_1.Buffer.alloc(bytesRead);
                        buf.copy(newBuf, 0, 0, bytesRead);
                        buf = newBuf;
                    }
                    this.onRead(buf, chState);
                }
                catch (e) {
                    this.agent.abort(e);
                }
            }
        });
    }
    closeChannel(chState) {
        bayLog_1.BayLog.debug("%s Close channel: %s", this.agent, chState);
        if (chState.ch.type == channelWrapper_1.ChannelWrapper.TYPE_SOCKET) {
            chState.ch.socket.end(); // 'close' event will occur
            try {
                this.onClosed(chState);
            }
            catch (e) {
                this.agent.abort(e);
            }
            this.removeChannelState(chState.ch);
        }
        else if (chState.ch.type == channelWrapper_1.ChannelWrapper.TYPE_READABLE) {
            chState.ch.readable.destroy(); // 'close' event will occur
        }
        else {
            fs.close(chState.ch.fd, (err) => {
                try {
                    if (err)
                        bayLog_1.BayLog.error_e(err);
                    if (chState.listener != null)
                        this.onClosed(chState);
                    this.removeChannelState(chState.ch);
                }
                catch (e) {
                    this.agent.abort(e);
                }
            });
        }
    }
}
exports.NonBlockingHandler = NonBlockingHandler;
//# sourceMappingURL=nonBlockingHandler.js.map