"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BuiltInLogDocker = void 0;
const dockerBase_1 = require("../base/dockerBase");
const logBoat_1 = require("./logBoat");
const bayLog_1 = require("../../bayLog");
const fs = require("fs");
const plainTransporter_1 = require("../../agent/transporter/plainTransporter");
const grandAgent_1 = require("../../agent/grandAgent");
const bayMessage_1 = require("../../bayMessage");
const symbol_1 = require("../../symbol");
const os = require("os");
const logItems_1 = require("./logItems");
const configException_1 = require("../../configException");
const sysUtil_1 = require("../../util/sysUtil");
const bayserver_1 = require("../../bayserver");
const path = require("path");
const channelWrapper_1 = require("../../agent/channelWrapper");
class AgentListener {
    constructor(dkr) {
        this.docker = dkr;
    }
    add(agtId) {
        let fileName = this.docker.filePrefix + "_" + agtId + "." + this.docker.fileExt;
        let boat = new logBoat_1.LogBoat();
        fs.open(fileName, "a", (err, fd) => {
            if (err) {
                bayLog_1.BayLog.error_e(err, "Cannot open file: %s", fileName);
                return;
            }
            bayLog_1.BayLog.debug("Log file opened: %s fd=%d", fileName, fd);
            let tp;
            if (this.docker.logWriteMethod == BuiltInLogDocker.LOG_WRITE_METHOD_SELECT) {
                tp = new plainTransporter_1.PlainTransporter(false, 0, true); // write only
                tp.init(grandAgent_1.GrandAgent.get(agtId).nonBlockingHandler, new channelWrapper_1.ChannelWrapper(fd), boat);
            }
            else {
                throw new Error("Log Write method not supported");
            }
            try {
                boat.initBoat(fileName, tp);
            }
            catch (e) {
                bayLog_1.BayLog.fatal(bayMessage_1.BayMessage.get(symbol_1.Symbol.INT_CANNOT_OPEN_LOG_FILE, fileName));
                bayLog_1.BayLog.fatal_e(e);
            }
        });
        this.docker.loggers.set(agtId, boat);
    }
    remove(agtId) {
        this.docker.loggers.delete(agtId);
    }
}
class BuiltInLogDocker extends dockerBase_1.DockerBase {
    constructor() {
        super();
        /** Log items */
        this.logItems = [];
        /** Log write method */
        this.logWriteMethod = BuiltInLogDocker.DEFAULT_LOG_WRITE_METHOD;
        this.loggers = new Map();
    }
    ////////////////////////////////////////
    // Implements DockerBase
    ////////////////////////////////////////
    init(elm, parent) {
        super.init(elm, parent);
        let p = elm.arg.lastIndexOf('.');
        if (p == -1) {
            this.filePrefix = elm.arg;
            this.fileExt = "";
        }
        else {
            this.filePrefix = elm.arg.substring(0, p);
            this.fileExt = elm.arg.substring(p + 1);
        }
        if (this.format == null) {
            throw new configException_1.ConfigException(elm.fileName, elm.lineNo, bayMessage_1.BayMessage.get(symbol_1.Symbol.CFG_INVALID_LOG_FORMAT, ""));
        }
        if (!sysUtil_1.SysUtil.isAbsolutePath(this.filePrefix))
            this.filePrefix = sysUtil_1.SysUtil.joinPath(bayserver_1.BayServer.bservHome, this.filePrefix);
        let logDir = path.dirname(this.filePrefix);
        if (!sysUtil_1.SysUtil.isDirectory(logDir))
            fs.mkdir(logDir, err => {
                bayLog_1.BayLog.fatal_e(err, "Cannot create log directory: %s", logDir);
            });
        // Parse format
        this.compile(this.format, this.logItems, elm.fileName, elm.lineNo);
        grandAgent_1.GrandAgent.addLifecycleListener(new AgentListener(this));
    }
    initKeyVal(kv) {
        switch (kv.key.toLowerCase()) {
            default:
                return false;
            case "format":
                this.format = kv.value;
                break;
            case "logwritemethod":
                switch (kv.value.toLowerCase()) {
                    case "select":
                        this.logWriteMethod = BuiltInLogDocker.LOG_WRITE_METHOD_SELECT;
                        break;
                    case "spin":
                        this.logWriteMethod = BuiltInLogDocker.LOG_WRITE_METHOD_SPIN;
                        break;
                    case "taxi":
                        this.logWriteMethod = BuiltInLogDocker.LOG_WRITE_METHOD_TAXI;
                        break;
                    default:
                        throw new configException_1.ConfigException(kv.fileName, kv.lineNo, bayMessage_1.BayMessage.get(symbol_1.Symbol.CFG_INVALID_PARAMETER_VALUE, kv.value));
                }
        }
        return true;
    }
    ////////////////////////////////////////
    // Implements Log
    ////////////////////////////////////////
    log(tour) {
        let sb = "";
        for (const logItem of this.logItems) {
            let item = logItem.getItem(tour);
            if (item == null)
                sb += "-";
            else
                sb += item;
        }
        // If threre are message to write, write it
        if (sb.length > 0) {
            try {
                this.getLogger(tour.ship.agent).log(sb.toString());
            }
            catch (e) {
                bayLog_1.BayLog.error_e(e, "%s Error on logging: %s", tour, e.message);
            }
        }
    }
    ////////////////////////////////////////
    // private methods
    ////////////////////////////////////////
    /**
     * Compile format pattern
     */
    compile(str, items, fileName, lineNo) {
        // Find control code
        let pos = str.indexOf('%');
        if (pos != -1) {
            let text = str.substring(0, pos);
            items.push(new logItems_1.TextItem(text));
            this.compileCtl(str.substring(pos + 1), items, fileName, lineNo);
        }
        else {
            items.push(new logItems_1.TextItem(str));
        }
    }
    /**
     * Compile format pattern(Control code)
     */
    compileCtl(str, items, fileName, lineNo) {
        let param = null;
        // if exists param
        if (str.charAt(0) == '{') {
            // find close bracket
            let pos = str.indexOf('}');
            if (pos == -1) {
                throw new configException_1.ConfigException(fileName, lineNo, bayMessage_1.BayMessage.get(symbol_1.Symbol.CFG_INVALID_LOG_FORMAT, this.format));
            }
            param = str.substring(1, pos);
            str = str.substring(pos + 1);
        }
        let ctlChar = "";
        let error = false;
        if (str.length == 0)
            error = true;
        if (!error) {
            // get control char
            ctlChar = str.substring(0, 1);
            str = str.substring(1);
            if (ctlChar == ">") {
                if (str.length == 0) {
                    error = true;
                }
                else {
                    ctlChar = str.substring(0, 1);
                    str = str.substring(1);
                }
            }
        }
        let fct = null;
        if (!error) {
            fct = BuiltInLogDocker.map.get(ctlChar);
            if (fct == null)
                error = true;
        }
        if (error) {
            throw new configException_1.ConfigException(fileName, lineNo, bayMessage_1.BayMessage.get(symbol_1.Symbol.CFG_INVALID_LOG_FORMAT, this.format + " (unknown control code: '%" + ctlChar + "')"));
        }
        let item = fct();
        item.init(param);
        items.push(item);
        this.compile(str, items, fileName, lineNo);
    }
    getLogger(agt) {
        return this.loggers.get(agt.agentId);
    }
    ////////////////////////////////////////
    // Implements Log
    ////////////////////////////////////////
    static initClass() {
        this.map = new Map();
        // Create mapping table
        this.map.set("a", logItems_1.RemoteHostItem.factory);
        this.map.set("A", logItems_1.ServerIpItem.factory);
        this.map.set("b", logItems_1.RequestBytesItem1.factory);
        this.map.set("B", logItems_1.RequestBytesItem2.factory);
        this.map.set("c", logItems_1.ConnectionStatusItem.factory);
        this.map.set("e", logItems_1.NullItem.factory);
        this.map.set("h", logItems_1.RemoteHostItem.factory);
        this.map.set("H", logItems_1.ProtocolItem.factory);
        this.map.set("i", logItems_1.RequestHeaderItem.factory);
        this.map.set("l", logItems_1.RemoteLogItem.factory);
        this.map.set("m", logItems_1.MethodItem.factory);
        this.map.set("n", logItems_1.NullItem.factory);
        this.map.set("o", logItems_1.ResponseHeaderItem.factory);
        this.map.set("p", logItems_1.PortItem.factory);
        this.map.set("P", logItems_1.NullItem.factory);
        this.map.set("q", logItems_1.QueryStringItem.factory);
        this.map.set("r", logItems_1.StartLineItem.factory);
        this.map.set("s", logItems_1.StatusItem.factory);
        this.map.set(">s", logItems_1.StatusItem.factory);
        this.map.set("t", logItems_1.TimeItem.factory);
        this.map.set("T", logItems_1.IntervalItem.factory);
        this.map.set("u", logItems_1.RemoteUserItem.factory);
        this.map.set("U", logItems_1.RequestUrlItem.factory);
        this.map.set("v", logItems_1.ServerNameItem.factory);
        this.map.set("V", logItems_1.NullItem.factory);
    }
}
exports.BuiltInLogDocker = BuiltInLogDocker;
BuiltInLogDocker.LOG_WRITE_METHOD_SELECT = 1;
BuiltInLogDocker.LOG_WRITE_METHOD_SPIN = 2;
BuiltInLogDocker.LOG_WRITE_METHOD_TAXI = 3;
BuiltInLogDocker.DEFAULT_LOG_WRITE_METHOD = BuiltInLogDocker.LOG_WRITE_METHOD_SELECT;
BuiltInLogDocker.lineSep = os.EOL;
BuiltInLogDocker.initClass();
module.exports = {
    createDocker: () => new BuiltInLogDocker()
};
//# sourceMappingURL=builtInLogDocker.js.map