"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BuiltInPermissionDocker = void 0;
const hostMatcher_1 = require("../../util/hostMatcher");
const ipMatcher_1 = require("../../util/ipMatcher");
const groups_1 = require("../../util/groups");
const dockerBase_1 = require("../base/dockerBase");
const configException_1 = require("../../configException");
const bayMessage_1 = require("../../bayMessage");
const symbol_1 = require("../../symbol");
const strUtil_1 = require("../../util/strUtil");
const bayLog_1 = require("../../bayLog");
const httpException_1 = require("../../httpException");
const httpStatus_1 = require("../../util/httpStatus");
const httpHeaders_1 = require("../../util/httpHeaders");
class CheckItem {
    constructor(matcher, admit) {
        this.matcher = matcher;
        this.admit = admit;
    }
    socketAdmitted(ch) {
        return this.matcher.matchSocket(ch) == this.admit;
    }
    tourAdmitted(tur) {
        return this.matcher.matchTour(tur) == this.admit;
    }
}
class HostPermissionMatcher {
    constructor(hostPtn) {
        this.mch = new hostMatcher_1.HostMatcher(hostPtn);
    }
    matchSocket(ch) {
        return this.mch.match(ch.remoteAddress);
    }
    matchTour(tur) {
        return this.mch.match(tur.req.remoteHost());
    }
}
class IpPermissionMatcher {
    constructor(ipDesc) {
        this.mch = new ipMatcher_1.IpMatcher(ipDesc);
    }
    matchSocket(ch) {
        return this.mch.match(ch.remoteAddress);
    }
    matchTour(tur) {
        return this.mch.match(tur.req.remoteAddress);
    }
}
class BuiltInPermissionDocker extends dockerBase_1.DockerBase {
    constructor() {
        super(...arguments);
        this.checkList = [];
        this.groups = [];
    }
    //////////////////////////////////////////////////////
    // Implements DockerBase
    //////////////////////////////////////////////////////
    initKeyVal(kv) {
        try {
            switch (kv.key.toLowerCase()) {
                default:
                    return false;
                case "admit":
                case "allow": {
                    for (const pm of this.parseValue(kv)) {
                        this.checkList.push(new CheckItem(pm, true));
                    }
                    break;
                }
                case "refuse":
                case "deny": {
                    for (const pm of this.parseValue(kv)) {
                        this.checkList.push(new CheckItem(pm, false));
                    }
                    break;
                }
                case "group": {
                    let groupList = kv.value.split((" "));
                    for (const gName of groupList) {
                        let g = groups_1.Groups.getGroup(gName);
                        if (g == null) {
                            throw new configException_1.ConfigException(kv.fileName, kv.lineNo, bayMessage_1.BayMessage.get(symbol_1.Symbol.CFG_GROUP_NOT_FOUND, kv.value));
                        }
                        this.groups.push(g);
                    }
                    break;
                }
            }
            return true;
        }
        catch (e) {
            if (e instanceof configException_1.ConfigException)
                throw e;
            else {
                bayLog_1.BayLog.error_e(e);
                throw new configException_1.ConfigException(kv.fileName, kv.lineNo, bayMessage_1.BayMessage.get(symbol_1.Symbol.CFG_INVALID_PERMISSION_DESCRIPTION, kv.value), e.message);
            }
        }
    }
    //////////////////////////////////////////////////////
    // Implements Permission
    //////////////////////////////////////////////////////
    socketAdmitted(ch) {
        // Check remote host
        let isOk = true;
        for (const chk of this.checkList) {
            if (chk.admit) {
                if (chk.socketAdmitted(ch)) {
                    isOk = true;
                    break;
                }
            }
            else {
                if (!chk.socketAdmitted(ch)) {
                    isOk = false;
                    break;
                }
            }
        }
        if (!isOk) {
            bayLog_1.BayLog.error("Permission error: socket not admitted: %s", ch);
            throw new httpException_1.HttpException(httpStatus_1.HttpStatus.FORBIDDEN);
        }
    }
    tourAdmitted(tour) {
        // Check remote host
        let isOk = true;
        for (const chk of this.checkList) {
            if (chk.admit) {
                if (chk.tourAdmitted(tour)) {
                    isOk = true;
                    break;
                }
            }
            else {
                if (!chk.tourAdmitted(tour)) {
                    isOk = false;
                    break;
                }
            }
        }
        if (!isOk)
            throw new httpException_1.HttpException(httpStatus_1.HttpStatus.FORBIDDEN, tour.req.uri);
        if (this.groups.length == 0)
            return;
        // Check member
        isOk = false;
        if (tour.req.remoteUser != null) {
            for (const g of this.groups) {
                if (g.validate(tour.req.remoteUser, tour.req.remotePass)) {
                    isOk = true;
                    break;
                }
            }
        }
        if (!isOk) {
            tour.res.headers.set(httpHeaders_1.HttpHeaders.WWW_AUTHENTICATE, "Basic realm=\"Auth\"");
            throw new httpException_1.HttpException(httpStatus_1.HttpStatus.UNAUTHORIZED);
        }
    }
    //////////////////////////////////////////////////////
    // private methods
    //////////////////////////////////////////////////////
    parseValue(kv) {
        let valList = kv.value.split(" ");
        let type = null;
        let matchStr = [];
        if (valList.length > 0) {
            type = valList.shift();
            matchStr = valList;
        }
        if (matchStr.length == 0) {
            throw new configException_1.ConfigException(kv.fileName, kv.lineNo, bayMessage_1.BayMessage.get(symbol_1.Symbol.CFG_INVALID_PERMISSION_DESCRIPTION, kv.value));
        }
        let pmList = [];
        if (strUtil_1.StrUtil.eqIgnoreCase(type, "host")) {
            for (const m of matchStr) {
                pmList.push(new HostPermissionMatcher(m));
            }
            return pmList;
        }
        else if (strUtil_1.StrUtil.eqIgnoreCase(type, "ip")) {
            for (const m of matchStr) {
                pmList.push(new IpPermissionMatcher(m));
            }
            return pmList;
        }
        else {
            throw new configException_1.ConfigException(kv.fileName, kv.lineNo, bayMessage_1.BayMessage.get(symbol_1.Symbol.CFG_INVALID_PERMISSION_DESCRIPTION, kv.value));
        }
    }
}
exports.BuiltInPermissionDocker = BuiltInPermissionDocker;
module.exports = {
    createDocker: () => new BuiltInPermissionDocker()
};
//# sourceMappingURL=builtInPermissionDocker.js.map