"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.WarpShip = exports.WarpShip_TourWrapper = void 0;
const ship_1 = require("../../watercraft/ship");
const tour_1 = require("../../tour/tour");
const bayserver_1 = require("../../bayserver");
const bayLog_1 = require("../../bayLog");
const warpData_1 = require("./warpData");
const sink_1 = require("../../sink");
const httpStatus_1 = require("../../util/httpStatus");
class WarpShip_TourWrapper {
    constructor(id, tour) {
        this.id = id;
        this.tour = tour;
    }
    toString() {
        return "<" + this.id + "," + this.tour.toString() + ">";
    }
}
exports.WarpShip_TourWrapper = WarpShip_TourWrapper;
class WarpShip extends ship_1.Ship {
    constructor() {
        super();
        this.cmdBuf = [];
        this.tourMap = new Map();
    }
    initWarp(ch, agt, tp, dkr, protoHandler) {
        super.init(ch, agt, tp);
        this.docker = dkr;
        this.socketTimeoutSec = dkr.timeoutSec >= 0 ? dkr.timeoutSec : bayserver_1.BayServer.harbor.getSocketTimeoutSec();
        this.setProtocolHandler(protoHandler);
    }
    toString() {
        return this.agent + " wsip#" + this.shipId + "/" + this.objectId + "[" + this.protocol() + "]";
    }
    //////////////////////////////////////////////////////
    // implements Reusable
    //////////////////////////////////////////////////////
    reset() {
        super.reset();
        if (this.tourMap.size != 0)
            bayLog_1.BayLog.error("BUG: %s Some tours is active: %s", this, this.tourMap);
        this.connected = false;
        this.tourMap.clear();
        this.cmdBuf.length = 0;
    }
    //////////////////////////////////////////////////////
    // Custom methods
    //////////////////////////////////////////////////////
    getDocker() {
        return this.docker;
    }
    getWarpHandler() {
        return this.protocolHandler;
    }
    startWarpTour(tur) {
        let wHnd = this.getWarpHandler();
        let warpId = wHnd.nextWarpId();
        let wdat = wHnd.newWarpData(warpId);
        bayLog_1.BayLog.debug("%s new warp tour related to %s", wdat, tur);
        tur.req.setContentHandler(wdat);
        bayLog_1.BayLog.debug("%s start: warpId=%d", wdat, warpId);
        if (this.tourMap.has(warpId))
            throw new sink_1.Sink("warpId exists");
        this.tourMap.set(warpId, new WarpShip_TourWrapper(tur.id(), tur));
        wHnd.postWarpHeaders(tur);
        if (this.connected) {
            bayLog_1.BayLog.debug("%s is already connected. Just start warp tour:%s", wdat, tur);
            wdat.start();
        }
    }
    endWarpTour(tur) {
        let wdat = warpData_1.WarpData.get(tur);
        bayLog_1.BayLog.debug("wrp#%d end (tur=%s): started=%b ended=%b", wdat.warpId, tur, wdat.started, wdat.ended);
        if (!this.tourMap.has(wdat.warpId))
            throw new sink_1.Sink("%s WarpId not in tourMap: %d", tur, wdat.warpId);
        else
            this.tourMap.delete(wdat.warpId);
        this.docker.keepShip(this);
    }
    notifyServiceUnavailable(msg) {
        this.notifyErrorToOwnerTour(httpStatus_1.HttpStatus.SERVICE_UNAVAILABLE, msg);
    }
    getTour(warpId, must = true) {
        let wpr = this.tourMap.get(warpId);
        if (wpr != null) {
            wpr.tour.checkTourId(wpr.id);
            if (!warpData_1.WarpData.get(wpr.tour).ended) {
                return wpr.tour;
            }
        }
        if (must)
            throw new sink_1.Sink("%s warp tours not found: id=%d", this, warpId);
        else
            return null;
    }
    //////////////////////////////////////////////////////
    // Other methods
    //////////////////////////////////////////////////////
    abort(checkId) {
        this.checkShipId(checkId);
        this.postman.abort();
    }
    isTimeout(duration) {
        var timeout;
        if (this.keeping) {
            // warp connection never timeout in keeping
            timeout = false;
        }
        else if (this.socketTimeoutSec <= 0)
            timeout = false;
        else
            timeout = duration >= this.socketTimeoutSec;
        bayLog_1.BayLog.debug(this + " Warp check timeout: dur=" + duration + ", timeout=" + timeout + ", keeping=" + this.keeping + " limit=" + this.socketTimeoutSec);
        return timeout;
    }
    notifyErrorToOwnerTour(status, msg) {
        for (const warpId of this.tourMap.keys()) {
            let tur = this.getTour(warpId);
            bayLog_1.BayLog.debug("%s send error to owner: %s running=%b", this, tur, tur.isRunning());
            if (tur.isRunning()) {
                try {
                    tur.res.sendError(tour_1.Tour.TOUR_ID_NOCHECK, status, msg);
                }
                catch (e) {
                    bayLog_1.BayLog.error_e(e);
                }
            }
            else {
                tur.res.endContent(tour_1.Tour.TOUR_ID_NOCHECK);
            }
        }
        this.tourMap = new Map();
    }
    endShip() {
        this.docker.returnProtocolHandler(this.agent, this.protocolHandler);
        this.docker.returnShip(this);
    }
    post(cmd, listener = null) {
        if (!this.connected)
            this.cmdBuf.push([cmd, listener]);
        else {
            if (cmd == null)
                listener();
            else
                this.protocolHandler.commandPacker.post(this, cmd, listener);
        }
    }
    flush() {
        for (const cmdAndLis of this.cmdBuf) {
            let cmd = cmdAndLis[0];
            let lis = cmdAndLis[1];
            if (cmd == null)
                lis();
            else
                this.protocolHandler.commandPacker.post(this, cmd, lis);
        }
        this.cmdBuf = [];
    }
}
exports.WarpShip = WarpShip;
//# sourceMappingURL=warpShip.js.map