"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Tour = void 0;
const counter_1 = require("../util/counter");
const tourReq_1 = require("./tourReq");
const tourRes_1 = require("./tourRes");
const httpException_1 = require("../httpException");
const sink_1 = require("../sink");
const bayLog_1 = require("../bayLog");
const bayserver_1 = require("../bayserver");
const httpStatus_1 = require("../util/httpStatus");
class Tour {
    constructor() {
        this.req = new tourReq_1.TourReq(this);
        this.res = new tourRes_1.TourRes(this);
        this.state = Tour.STATE_UNINITIALIZED;
        this.objectId = Tour.oidCounter.next();
    }
    init(key, sip) {
        if (this.isInitialized())
            throw new sink_1.Sink("%s Tour already initialized (state=%d)", this, this.state);
        this.ship = sip;
        this.shipId = sip.id();
        this.tourId = Tour.idCounter.next();
        this.req.key = key;
        this.req.init(key);
        this.res.init();
        this.changeState(Tour.TOUR_ID_NOCHECK, Tour.STATE_PREPARING);
        bayLog_1.BayLog.debug(this + " initialized");
    }
    toString() {
        return this.ship + " tour#" + this.tourId + "/" + this.objectId + "[key=" + this.req.key + "]";
    }
    //////////////////////////////////////////////////////
    // Implements Reusable
    //////////////////////////////////////////////////////
    reset() {
        bayLog_1.BayLog.trace("%s reset", this);
        this.req.reset();
        this.res.reset();
        this.city = null;
        this.town = null;
        this.club = null;
        this.errorHandling = false;
        this.tourId = Tour.INVALID_TOUR_ID;
        this.interval = 0;
        this.isSecure = false;
        //BayLog.trace("%s reset running false", this);
        this.changeState(Tour.TOUR_ID_NOCHECK, Tour.STATE_UNINITIALIZED);
        this.error = null;
        this.ship = null;
    }
    //////////////////////////////////////////////////////
    // Other methods
    //////////////////////////////////////////////////////
    id() {
        return this.tourId;
    }
    go() {
        let city = this.ship.portDocker.findCity(this.req.reqHost);
        if (city == null)
            city = bayserver_1.BayServer.findCity(this.req.reqHost);
        this.changeState(Tour.TOUR_ID_NOCHECK, Tour.STATE_RUNNING);
        bayLog_1.BayLog.debug("%s GO TOUR! ...( ^_^)/: city=%s url=%s", this, this.req.reqHost, this.req.uri);
        if (city == null) {
            throw new httpException_1.HttpException(httpStatus_1.HttpStatus.NOT_FOUND, this.req.uri);
        }
        else {
            try {
                city.enter(this);
            }
            catch (e) {
                this.changeState(Tour.TOUR_ID_NOCHECK, Tour.STATE_ABORTED);
                if (e instanceof sink_1.Sink) {
                    throw e;
                }
                else if (e instanceof httpException_1.HttpException) {
                    bayLog_1.BayLog.error_e(e);
                    throw e;
                }
                else {
                    bayLog_1.BayLog.error_e(e);
                    throw new httpException_1.HttpException(httpStatus_1.HttpStatus.INTERNAL_SERVER_ERROR, e.message);
                }
            }
        }
    }
    isValid() {
        return this.state == Tour.STATE_PREPARING || this.state == Tour.STATE_RUNNING;
    }
    isPreparing() {
        return this.state == Tour.STATE_PREPARING;
    }
    isRunning() {
        return this.state == Tour.STATE_RUNNING;
    }
    isAborted() {
        return this.state == Tour.STATE_ABORTED;
    }
    isZombie() {
        return this.state == Tour.STATE_ZOMBIE;
    }
    isEnded() {
        return this.state == Tour.STATE_ENDED;
    }
    isInitialized() {
        return this.state != Tour.STATE_UNINITIALIZED;
    }
    changeState(checkId, newState) {
        bayLog_1.BayLog.trace("%s change state: %s", this, newState);
        this.checkTourId(checkId);
        this.state = newState;
    }
    checkTourId(checkId) {
        if (checkId == Tour.TOUR_ID_NOCHECK)
            return;
        if (!this.isInitialized()) {
            throw new sink_1.Sink("%s Tour not initialized", this);
        }
        if (checkId != this.tourId) {
            throw new sink_1.Sink("%s Invalid tour id : %d", this, this.tourId);
        }
    }
}
exports.Tour = Tour;
Tour.STATE_UNINITIALIZED = 0;
Tour.STATE_PREPARING = 1;
Tour.STATE_RUNNING = 2;
Tour.STATE_ABORTED = 3;
Tour.STATE_ENDED = 4;
Tour.STATE_ZOMBIE = 5;
Tour.TOUR_ID_NOCHECK = -1;
Tour.INVALID_TOUR_ID = 0;
Tour.oidCounter = new counter_1.Counter();
Tour.idCounter = new counter_1.Counter();
//# sourceMappingURL=tour.js.map