"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TourReq = void 0;
const tour_1 = require("./tour");
const httpHeaders_1 = require("../util/httpHeaders");
const sink_1 = require("../sink");
const bayLog_1 = require("../bayLog");
const protocolException_1 = require("../protocol/protocolException");
const bayserver_1 = require("../bayserver");
const bayMessage_1 = require("../bayMessage");
const symbol_1 = require("../symbol");
class TourReq {
    constructor(tour) {
        this.headers = new httpHeaders_1.HttpHeaders();
        this.tour = tour;
    }
    init(key) {
        this.key = key;
    }
    //////////////////////////////////////////////////////
    // Private methods
    //////////////////////////////////////////////////////
    reset() {
        this.headers.clear();
        this.key = 0;
        this.uri = null;
        this.method = null;
        this.protocol = null;
        this.bytesPosted = 0;
        this.bytesConsumed = 0;
        this.bytesLimit = 0;
        this.rewrittenURI = null;
        this.queryString = null;
        this.pathInfo = null;
        this.scriptName = null;
        this.reqHost = null;
        this.reqPort = 0;
        this.remoteUser = null;
        this.remotePass = null;
        this.remoteAddress = null;
        this.remotePort = 0;
        this.remoteHostFunc = null;
        this.serverAddress = null;
        this.serverPort = 0;
        this.serverName = null;
        this.charset = null;
        this.contentHandler = null;
        this.consumeListener = null;
        this.available = false;
        this.ended = false;
    }
    //////////////////////////////////////////////////////
    // Other methods
    //////////////////////////////////////////////////////
    remoteHost() {
        if (this.remoteHostFunc == null)
            return null;
        else
            return this.remoteHostFunc();
    }
    setContentHandler(hnd) {
        if (hnd == null)
            throw new Error("nullPo");
        if (this.contentHandler != null)
            throw new sink_1.Sink("content handler is already set");
        this.contentHandler = hnd;
    }
    setConsumeListener(limit, listener) {
        if (limit < 0) {
            throw new sink_1.Sink("Invalid limit");
        }
        this.consumeListener = listener;
        this.bytesLimit = limit;
        this.bytesConsumed = 0;
        this.bytesPosted = 0;
        this.available = true;
    }
    postContent(checkId, data, start, len) {
        this.tour.checkTourId(checkId);
        let dataPassed = false;
        if (!this.tour.isRunning()) {
            bayLog_1.BayLog.debug("%s tour is not running.", this.tour);
        }
        else if (this.tour.req.contentHandler == null) {
            bayLog_1.BayLog.warn("%s content read, but no content handler", this.tour);
        }
        else if (this.consumeListener == null) {
            throw new sink_1.Sink("Request consume listener is null");
        }
        else if (this.bytesPosted + len > this.bytesLimit) {
            throw new protocolException_1.ProtocolException(bayMessage_1.BayMessage.get(symbol_1.Symbol.HTP_READ_DATA_EXCEEDED, this.bytesPosted + len, this.bytesLimit));
        }
        else if (this.tour.error != null)
            // If has error, only read content. (Do not call content handler)
            bayLog_1.BayLog.debug("%s tour has error.", this.tour);
        else {
            this.contentHandler.onReadContent(this.tour, data, start, len);
            dataPassed = true;
        }
        this.bytesPosted += len;
        bayLog_1.BayLog.debug("%s read content: len=%d posted=%d limit=%d consumed=%d available=%s", this.tour, len, this.bytesPosted, this.bytesLimit, this.bytesConsumed, this.available);
        if (!dataPassed)
            return true;
        let oldAvailable = this.available;
        if (!this.bufferAvailable())
            this.available = false;
        if (oldAvailable && !this.available) {
            bayLog_1.BayLog.debug("%s request unavailable (_ _).zZZ: posted=%d consumed=%d", this, this.bytesPosted, this.bytesConsumed);
        }
        return this.available;
    }
    endContent(checkId) {
        bayLog_1.BayLog.debug(this.tour + " endReqContent");
        this.tour.checkTourId(checkId);
        if (this.ended)
            throw new sink_1.Sink(this.tour + " Request content is already ended");
        if (this.bytesLimit >= 0 && this.bytesPosted != this.bytesLimit) {
            throw new protocolException_1.ProtocolException("nvalid request data length: " + this.bytesPosted + "/" + this.bytesLimit);
        }
        if (this.contentHandler != null)
            this.contentHandler.onEndContent(this.tour);
        this.ended = true;
    }
    consumed(checkId, length) {
        this.tour.checkTourId(checkId);
        if (this.consumeListener == null)
            throw new sink_1.Sink("Request consume listener is null");
        this.bytesConsumed += length;
        bayLog_1.BayLog.debug("%s reqConsumed: len=%d posted=%d limit=%d consumed=%d available=%s", this.tour, length, this.bytesPosted, this.bytesLimit, this.bytesConsumed, this.available);
        let resume = false;
        let oldAvailable = this.available;
        if (this.bufferAvailable())
            this.available = true;
        if (!oldAvailable && this.available) {
            bayLog_1.BayLog.debug("%s request available (^o^): posted=%d consumed=%d", this, this.bytesPosted, this.bytesConsumed);
            resume = true;
        }
        this.consumeListener(length, resume);
    }
    abort() {
        bayLog_1.BayLog.debug("%s abort tour", this.tour);
        if (this.tour.isPreparing()) {
            this.tour.changeState(tour_1.Tour.TOUR_ID_NOCHECK, tour_1.Tour.STATE_ABORTED);
            return true;
        }
        else if (this.tour.isRunning()) {
            let aborted = true;
            if (this.contentHandler != null)
                aborted = this.contentHandler.onAbort(this.tour);
            if (aborted)
                this.tour.changeState(tour_1.Tour.TOUR_ID_NOCHECK, tour_1.Tour.STATE_ABORTED);
            return aborted;
        }
        else {
            bayLog_1.BayLog.debug("%s tour is already aborted", this.tour);
            return false;
        }
    }
    bufferAvailable() {
        return this.bytesPosted - this.bytesConsumed < bayserver_1.BayServer.harbor.getTourBufferSize();
    }
}
exports.TourReq = TourReq;
//# sourceMappingURL=tourReq.js.map